#![allow(dead_code)]

use image::{ImageBuffer, Rgb, ImageResult};
use imageproc::drawing;
use std::path::Path;

type ImgBufU8 = ImageBuffer<Rgb<u8>, Vec<u8>>;
type Colour = Rgb<u8>;
type Coord = (i32, i32);

pub mod defaults {
    use super::{Colour, Coord, Rgb};

    pub const WIDTH: u32 = 300;
    pub const HEIGHT: u32 = 300;
    pub const BACKGROUND_COLOUR: Colour = Rgb([255u8, 255u8, 255u8]);
    pub const COLOUR: Colour = Rgb([0u8, 0u8, 0u8]);
    pub const STEP_LENGTH: u32 = 10;
    pub const ANGLE: f64 = 0.0;

    pub fn position() -> Coord {
        ((WIDTH as f64 / 2.0) as i32, 
        (HEIGHT as f64 / 2.0) as i32)
    }
}

pub struct Turtle {
    buf: ImgBufU8,
    colour: Colour,
    pos: Coord,
    step_length: u32,
    angle: f64,
}

impl Turtle {
    pub fn new() -> Self {
        let mut buf = ImageBuffer::new(defaults::WIDTH, defaults::HEIGHT);
        for pixel in buf.pixels_mut() {
            *pixel = defaults::BACKGROUND_COLOUR;
        }

        Turtle {
            buf,
            colour: defaults::COLOUR,
            pos: defaults::position(),
            step_length: defaults::STEP_LENGTH,
            angle: defaults::ANGLE,
        }
    }

    pub fn step_length(&self) -> u32 {
        self.step_length
    }

    pub fn position(&self) -> Coord {
        self.pos
    }

    pub fn angle(&self) -> f64 {
        self.angle.to_degrees()
    }

    pub fn set_pos(&mut self, coord: Coord) {
        self.pos.0 = coord.0;
        self.pos.1 = coord.1;
    }

    pub fn right(&mut self, degrees: f64) {
        self.angle += degrees.to_radians();
        self.angle = self.angle.rem_euclid(2.0 * std::f64::consts::PI);
    }

    pub fn right_forward(&mut self, degrees: f64, steps_count: u32) {
        self.right(degrees);
        self.forward(steps_count);
    }

    pub fn forward_right(&mut self, steps_count: u32, degrees: f64) {
        self.forward(steps_count);
        self.right(degrees);
    }

    pub fn left(&mut self, degrees: f64) {
        self.angle -= degrees.to_radians();
        self.angle = self.angle.rem_euclid(2.0 * std::f64::consts::PI);
    }

    pub fn left_forward(&mut self, degrees: f64, steps_count: u32) {
        self.left(degrees);
        self.forward(steps_count);
    }

    pub fn forward_left(&mut self, steps_count: u32, degrees: f64) {
        self.forward(steps_count);
        self.left(degrees);
    }

    pub fn forward(&mut self, steps_count: u32) {
        let start_pos = self.pos;

        let dy = self.angle.sin() * steps_count as f64 * self.step_length as f64;
        let dx = self.angle.cos() * steps_count as f64 * self.step_length as f64;
        
        self.set_pos((
            self.pos.0 + dx.round() as i32,
            self.pos.1 + dy.round() as i32
        ));

        self.draw_line(start_pos, self.pos);
    }

    pub fn save<P>(&self, output_path: P) -> ImageResult<()> 
    where
        P: AsRef<Path>
    {
        self.buf.save(output_path)
    }

    fn draw_line(&mut self, start: Coord, end: Coord) {
        drawing::draw_line_segment_mut(
            &mut self.buf,
            (start.0 as f32, start.1 as f32),
            (end.0 as f32, end.1 as f32),
            self.colour
        );
    }
}